/*
  Name: a3Dmath.hpp
  Copyright: Copyright(C) 2004, Jan Ringos, Tringi@Mx-3.cz
  Author: Jan Ringos, http://Tringi.Mx-3.cz
  Description: a3Dmath vertor library
  
  This is not official version of the code.
*/
#include "a3DMath.hpp"

//	This returns a perpendicular vector from 2 given vectors by taking the cross product.
VECTOR3D Cross(VECTOR3D vVector1, VECTOR3D vVector2){
	return (VECTOR3D){((vVector1.y * vVector2.z) - (vVector1.z * vVector2.y)),
                      ((vVector1.z * vVector2.x) - (vVector1.x * vVector2.z)),
                      ((vVector1.x * vVector2.y) - (vVector1.y * vVector2.x)) };
}  
 
//	This returns a vector between 2 points
VECTOR3D Vector(VECTOR3D vPoint1, VECTOR3D vPoint2){
	return (VECTOR3D){vPoint1.x-vPoint2.x,vPoint1.y-vPoint2.y,vPoint1.z-vPoint2.z};
}

//	This returns the magnitude of a normal (or any other vector)
float Magnitude(VECTOR3D vNormal){
 return sqrt((vNormal.x*vNormal.x)+(vNormal.y*vNormal.y)+(vNormal.z*vNormal.z));
}

//	This returns a normalize vector (A vector exactly of length 1)
VECTOR3D Normalize(VECTOR3D vNormal){
    register double magnitude = Magnitude(vNormal);
	return (VECTOR3D){vNormal.x/=magnitude,vNormal.y/=magnitude,vNormal.z/=magnitude};
}

//	This returns the normal of a polygon (The direction the polygon is facing)
VECTOR3D Normal(VECTOR3D vTriangle[]){
	return Normalize( Cross( Vector(vTriangle[2], vTriangle[0]),
                             Vector(vTriangle[1], vTriangle[0])) );
}

VECTOR3D Normal3(VECTOR3D a, VECTOR3D b, VECTOR3D c){
	return Normalize( Cross(Vector(c,a), Vector(b,a)) );
}

float PlaneDistance(VECTOR3D Normal, VECTOR3D Point){	
	return - ((Normal.x * Point.x) + (Normal.y * Point.y) + (Normal.z * Point.z));
}

//	This checks to see if a line intersects a plane
bool IntersectedPlane(VECTOR3D vPoly[], VECTOR3D vLine[],
                      VECTOR3D &vNormal, float &originDistance){
	float distance1=0, distance2=0;
	vNormal = Normal(vPoly);
	originDistance = PlaneDistance(vNormal, vPoly[0]);
	distance1 = ((vNormal.x * vLine[0].x)  +
		         (vNormal.y * vLine[0].y)  +
				 (vNormal.z * vLine[0].z)) + originDistance;
	distance2 = ((vNormal.x * vLine[1].x)  +
		         (vNormal.y * vLine[1].y)  +
				 (vNormal.z * vLine[1].z)) + originDistance;
	if(distance1 * distance2 >= 0)
	   return false;

	return true;
}

//	This computes the dot product of 2 vectors
float Dot(VECTOR3D vVector1, VECTOR3D vVector2) {
 return((vVector1.x*vVector2.x)+(vVector1.y*vVector2.y)+(vVector1.z*vVector2.z));
}

//	This checks to see if a point is inside the ranges of a polygon
double AngleBetweenVectors(VECTOR3D Vector1, VECTOR3D Vector2){
	float dotProduct = Dot(Vector1, Vector2);				
	float vectorsMagnitude = Magnitude(Vector1) * Magnitude(Vector2) ;
	return( acos( dotProduct / vectorsMagnitude ) );
}

//	This returns the intersection point of the line that intersects the plane
VECTOR3D IntersectionPoint(VECTOR3D vNormal, VECTOR3D vLine[], double distance){
	VECTOR3D vPoint = {0}, vLineDir = {0};
	double Numerator = 0.0, Denominator = 0.0, dist = 0.0;
	vLineDir = Vector(vLine[1], vLine[0]);
	vLineDir = Normalize(vLineDir);
	Numerator = - (vNormal.x * vLine[0].x +
				   vNormal.y * vLine[0].y +
				   vNormal.z * vLine[0].z + distance);
	Denominator = Dot(vNormal, vLineDir);

	if( Denominator == 0.0)
		return vLine[0];
	dist = Numerator / Denominator;

	vPoint.x = (float)(vLine[0].x + (vLineDir.x * dist));
	vPoint.y = (float)(vLine[0].y + (vLineDir.y * dist));
	vPoint.z = (float)(vLine[0].z + (vLineDir.z * dist));
	return vPoint;
}

//	This returns the intersection distance of the line that intersects the plane
float IntersectionDist(VECTOR3D vNormal, VECTOR3D vLine[], double distance){
	VECTOR3D vLineDir = {0};
	double Numerator = 0.0, Denominator = 0.0, dist = 0.0;
	vLineDir = Vector(vLine[1], vLine[0]);
	vLineDir = Normalize(vLineDir);
	Numerator = - (vNormal.x * vLine[0].x +
				   vNormal.y * vLine[0].y +
				   vNormal.z * vLine[0].z + distance);
	Denominator = Dot(vNormal, vLineDir);

	if( Denominator == 0.0)
		return -1;
	dist = Numerator / Denominator;

	return dist;
}


//	This checks to see if a point is inside the ranges of a polygon
bool InsidePolygon(VECTOR3D vIntersection, VECTOR3D Poly[], long verticeCount){
	const double MATCH_FACTOR = 0.9999;
	double Angle = 0.0;
	VECTOR3D vA, vB;
	
	for (int i = 0; i < verticeCount; i++){	
		vA = Vector(Poly[i], vIntersection);
		vB = Vector(Poly[(i + 1) % verticeCount], vIntersection);
		Angle += AngleBetweenVectors(vA, vB);
	}

	if(Angle >= (MATCH_FACTOR * (2.0 * PI)) ) return true;
	return false;
}
 
float IntersectedPolygon(VECTOR3D vPoly[], VECTOR3D vLine[], int verticeCount){
	VECTOR3D vNormal = {0};
	float originDistance = 0;
	if(!IntersectedPlane(vPoly, vLine,   vNormal,   originDistance)) return 0;
	VECTOR3D vIntersection = IntersectionPoint(vNormal, vLine, originDistance);
	if(InsidePolygon(vIntersection, vPoly, verticeCount))
       return IntersectionDist(vNormal,vLine,originDistance);
	return 0; 
}

float PointDistance(VECTOR3D pt1, VECTOR3D pt2) {
    float dist = 0.0;
    float len[3];
    
    len[0] = (pt1.x - pt2.x);
    len[1] = (pt1.y - pt2.y);
    len[2] = (pt1.z - pt2.z);
    
    return sqrtf( len[0]*len[0] + len[1]*len[1] + len[2]*len[2] );
};

double PointDistance(DVECTOR3D pt1, DVECTOR3D pt2) {
    double dist = 0.0;
    double len[3];
    
    len[0] = (pt1.x - pt2.x);
    len[1] = (pt1.y - pt2.y);
    len[2] = (pt1.z - pt2.z);
    
    return sqrt( len[0]*len[0] + len[1]*len[1] + len[2]*len[2] );
};

